@@ -230,9 +230,8 @@ sub new {
   my $self = {
     'main'              => $main,
     'conf'		=> $main->{conf},
-    'log_raw_counts'	=> 0,
+    ## 'log_raw_counts'	=> 0, # see compute_prob_for_token()
     'use_ignores'       => 1,
-    'tz'		=> Mail::SpamAssassin::Util::local_tz(),
   };
   bless ($self, $class);
 
@@ -263,7 +262,7 @@ sub finish {
   # use Carp qw(cluck); cluck "stack trace at untie";
 
   $self->{store}->untie_db();
-  delete $self->{store};
+  %{$self} = ();
 }
 
 sub sa_die { Mail::SpamAssassin::sa_die(@_); }
@@ -382,19 +381,10 @@ sub tokenize_line {
 
     # but extend the stop-list. These are squarely in the gray
     # area, and it just slows us down to record them.
+    # See http://wiki.apache.org/spamassassin/BayesStopList for more info.
+    #
     next if $len < 3 ||
-	($token =~ /^(?:a(?:nd|ny|ble|ll|re)|
-		m(?:uch|ost|ade|ore|ail|ake|ailing|any|ailto)|
-		t(?:his|he|ime|hrough|hat)|
-		w(?:hy|here|ork|orld|ith|ithout|eb)|
-		f(?:rom|or|ew)| e(?:ach|ven|mail)|
-		o(?:ne|ff|nly|wn|ut)| n(?:ow|ot|eed)|
-		s(?:uch|ame)| l(?:ook|ike|ong)|
-		y(?:ou|our|ou're)|
-		The|has|have|into|using|http|see|It's|it's|
-		number|just|both|come|years|right|know|already|
-		people|place|first|because|
-		And|give|year|information|can)$/x);
+	($token =~ /^(?:a(?:ble|l(?:ready|l)|n[dy]|re)|b(?:ecause|oth)|c(?:an|ome)|e(?:ach|mail|ven)|f(?:ew|irst|or|rom)|give|h(?:a(?:ve|s)|ttp)|i(?:n(?:formation|to)|t\'s)|just|know|l(?:ike|o(?:ng|ok))|m(?:a(?:de|il(?:(?:ing|to))?|ke|ny)|o(?:re|st)|uch)|n(?:eed|o[tw]|umber)|o(?:ff|n(?:ly|e)|ut|wn)|p(?:eople|lace)|right|s(?:ame|ee|uch)|t(?:h(?:at|is|rough|e)|ime)|using|w(?:eb|h(?:ere|y)|ith(?:out)?|or(?:ld|k))|y(?:ears?|ou(?:(?:\'re|r))?))$/i);
 
     # are we in the body?  If so, apply some body-specific breakouts
     if ($region == 1 || $region == 2) {
@@ -673,8 +663,10 @@ sub ignore_message {
 
   return 0 unless $self->{use_ignores};
 
-  my $ignore = $PMS->check_from_in_list('bayes_ignore_from')
-    		|| $PMS->check_to_in_list('bayes_ignore_to');
+  my $ig_from = $self->{main}->call_plugins ("check_wb_list", { permsgstatus => $PMS, type => 'from', list => 'bayes_ignore_from' });
+  my $ig_to = $self->{main}->call_plugins ("check_wb_list", { permsgstatus => $PMS, type => 'to', list => 'bayes_ignore_to' });
+
+  my $ignore = $ig_from || $ig_to;
 
   dbg("bayes: not using bayes, bayes_ignore_from or _to rule") if $ignore;
 
@@ -976,9 +968,9 @@ sub get_body_from_msg {
     return { };
   }
 
-  $msg->extract_message_metadata ($self->{main});
   my $permsgstatus =
         Mail::SpamAssassin::PerMsgStatus->new($self->{main}, $msg);
+  $msg->extract_message_metadata ($permsgstatus);
   my $msgdata = $self->get_msgdata_from_permsgstatus ($permsgstatus);
   $permsgstatus->finish();
 
@@ -1066,9 +1058,15 @@ sub compute_prob_for_token {
             ($Mail::SpamAssassin::Bayes::Combine::FW_S_CONSTANT + $robn);
   }
 
-  if ($self->{log_raw_counts}) {
-    $self->{raw_counts} .= " s=$s,n=$n ";
-  }
+  # 'log_raw_counts' is used to log the raw data for the Bayes equations during
+  # a mass-check, allowing the S and X constants to be optimized quickly
+  # without requiring re-tokenization of the messages for each attempt. There's
+  # really no need for this code to be uncommented in normal use, however.   It
+  # has never been publicly documented, so commenting it out is fine. ;)
+
+  ## if ($self->{log_raw_counts}) {
+  ## $self->{raw_counts} .= " s=$s,n=$n ";
+  ## }
 
   return $prob;
 }
@@ -1174,9 +1172,9 @@ sub scan {
 
   my ($ns, $nn) = $self->{store}->nspam_nham_get();
 
-  if ($self->{log_raw_counts}) {
-    $self->{raw_counts} = " ns=$ns nn=$nn ";
-  }
+  ## if ($self->{log_raw_counts}) { # see compute_prob_for_token()
+  ## $self->{raw_counts} = " ns=$ns nn=$nn ";
+  ## }
 
   dbg("bayes: corpus size: nspam = $ns, nham = $nn");
 
@@ -1186,17 +1184,19 @@ sub scan {
 
   my $tokensdata = $self->{store}->tok_get_all(keys %{$msgtokens});
 
-  my %pw;
+  my %pw = ();
 
   foreach my $tokendata (@{$tokensdata}) {
     my ($token, $tok_spam, $tok_ham, $atime) = @{$tokendata};
     my $prob = $self->compute_prob_for_token($token, $ns, $nn, $tok_spam, $tok_ham);
-    if (defined($prob)) {
-      $pw{$token}->{prob} = $prob;
-      $pw{$token}->{spam_count} = $tok_spam;
-      $pw{$token}->{ham_count} = $tok_ham;
-      $pw{$token}->{atime} = $atime;
-    }
+    next unless defined $prob;
+
+    $pw{$token} = {
+      prob => $prob,
+      spam_count => $tok_spam,
+      ham_count => $tok_ham,
+      atime => $atime
+    };
   }
 
   # If none of the tokens were found in the DB, we're going to skip
@@ -1222,38 +1222,44 @@ sub scan {
   my $count = N_SIGNIFICANT_TOKENS;
   my @sorted = ();
 
-  my ($tcount_spammy,$tcount_hammy) = (0,0);
+  my @touch_tokens;
   my $tinfo_spammy = $permsgstatus->{bayes_token_info_spammy} = [];
   my $tinfo_hammy = $permsgstatus->{bayes_token_info_hammy} = [];
 
-  my @touch_tokens;
+  my %tok_strength = map { $_ => (abs($pw{$_}->{prob} - 0.5)) } keys %pw;
+  my $log_each_token = (would_log('dbg', 'bayes') > 1);
 
-  for (sort {
-              abs($pw{$b}->{prob} - 0.5) <=> abs($pw{$a}->{prob} - 0.5)
+  foreach my $tok (sort {
+              $tok_strength{$b} <=> $tok_strength{$a}
             } keys %pw)
   {
     if ($count-- < 0) { last; }
-    my $pw = $pw{$_}->{prob};
-    next if (abs($pw - 0.5) < 
+    next if ($tok_strength{$tok} <
                 $Mail::SpamAssassin::Bayes::Combine::MIN_PROB_STRENGTH);
 
+    my $pw = $pw{$tok}->{prob};
+
     # What's more expensive, scanning headers for HAMMYTOKENS and
     # SPAMMYTOKENS tags that aren't there or collecting data that
     # won't be used?  Just collecting the data is certainly simpler.
     #
-    my $raw_token = $msgtokens->{$_} || "(unknown)";
-    my $s = $pw{$_}->{spam_count};
-    my $n = $pw{$_}->{ham_count};
-    my $a = $pw{$_}->{atime};
-    push @$tinfo_spammy, [$raw_token,$pw,$s,$n,$a] if $pw >= 0.5 && ++$tcount_spammy;
-    push @$tinfo_hammy,  [$raw_token,$pw,$s,$n,$a] if $pw <  0.5 && ++$tcount_hammy;
+    my $raw_token = $msgtokens->{$tok} || "(unknown)";
+    my $s = $pw{$tok}->{spam_count};
+    my $n = $pw{$tok}->{ham_count};
+    my $a = $pw{$tok}->{atime};
+
+    if ($pw < 0.5) {
+      push @$tinfo_hammy,  [$raw_token,$pw,$s,$n,$a];
+    } else {
+      push @$tinfo_spammy, [$raw_token,$pw,$s,$n,$a];
+    }
 
     push (@sorted, $pw);
 
     # update the atime on this token, it proved useful
-    push(@touch_tokens, $_);
+    push(@touch_tokens, $tok);
 
-    if (would_log('dbg', 'bayes') > 1) {
+    if ($log_each_token) {
       dbg("bayes: token '$raw_token' => $pw");
     }
   }
@@ -1280,9 +1286,9 @@ sub scan {
   $permsgstatus->{bayes_nspam} = $ns;
   $permsgstatus->{bayes_nham} = $nn;
 
-  if ($self->{log_raw_counts}) {
-    print "#Bayes-Raw-Counts: $self->{raw_counts}\n";
-  }
+  ## if ($self->{log_raw_counts}) { # see compute_prob_for_token()
+  ## print "#Bayes-Raw-Counts: $self->{raw_counts}\n";
+  ## }
 
   $self->{main}->call_plugins("bayes_scan", { toksref => $msgtokens,
 					      probsref => \%pw,
@@ -1318,8 +1324,10 @@ skip:
     $self->{store}->untie_db();
   }
 
-  $permsgstatus->{tag_data}{BAYESTCHAMMY} = $tcount_hammy;
-  $permsgstatus->{tag_data}{BAYESTCSPAMMY} = $tcount_spammy;
+  $permsgstatus->{tag_data}{BAYESTCHAMMY} = 
+                        ($tinfo_hammy ? scalar @{$tinfo_hammy} : 0);
+  $permsgstatus->{tag_data}{BAYESTCSPAMMY} = 
+                        ($tinfo_spammy ? scalar @{$tinfo_spammy} : 0);
   $permsgstatus->{tag_data}{BAYESTCLEARNED} = $tcount_learned;
   $permsgstatus->{tag_data}{BAYESTC} = $tcount_total;
 
